@@ -61,9 +61,6 @@ using json = nlohmann::json;
 // Nodes are considered active if they've queried in less than this long
 #define ZT_NETCONF_NODE_ACTIVE_THRESHOLD (ZT_NETWORK_AUTOCONF_DELAY * 2)
 
-// Timeout for disk read cache (ms)
-#define ZT_NETCONF_DB_CACHE_TTL 60000
-
 namespace ZeroTier {
 
 static json _renderRule(ZT_VirtualNetworkRule &rule)
@@ -432,12 +429,11 @@ static bool _parseRule(json &r,ZT_VirtualNetworkRule &rule)
 }
 
 EmbeddedNetworkController::EmbeddedNetworkController(Node *node,const char *dbPath) :
+	_startTime(OSUtils::now()),
 	_threadsStarted(false),
 	_db(dbPath),
 	_node(node)
 {
-	OSUtils::mkdir(dbPath);
-	OSUtils::lockDownFile(dbPath,true); // networks might contain auth tokens, etc., so restrict directory permissions
 }
 
 EmbeddedNetworkController::~EmbeddedNetworkController()
@@ -503,7 +499,7 @@ unsigned int EmbeddedNetworkController::handleControlPlaneHttpGET(
 			json network;
 			{
 				Mutex::Lock _l(_db_m);
-				network = _db.get("network",nwids,ZT_NETCONF_DB_CACHE_TTL);
+				network = _db.get("network",nwids);
 			}
 			if (!network.size())
 				return 404;
@@ -518,7 +514,7 @@ unsigned int EmbeddedNetworkController::handleControlPlaneHttpGET(
 						json member;
 						{
 							Mutex::Lock _l(_db_m);
-							member = _db.get("network",nwids,"member",Address(address).toString(),ZT_NETCONF_DB_CACHE_TTL);
+							member = _db.get("network",nwids,"member",Address(address).toString());
 						}
 						if (!member.size())
 							return 404;
@@ -533,7 +529,7 @@ unsigned int EmbeddedNetworkController::handleControlPlaneHttpGET(
 						Mutex::Lock _l(_db_m);
 
 						responseBody = "{";
-						_db.filter((std::string("network/") + nwids + "/member/"),ZT_NETCONF_DB_CACHE_TTL,[&responseBody](const std::string &n,const json &member) {
+						_db.filter((std::string("network/") + nwids + "/member/"),[&responseBody](const std::string &n,const json &member) {
 							if ((member.is_object())&&(member.size() > 0)) {
 								responseBody.append((responseBody.length() == 1) ? "\"" : ",\"");
 								responseBody.append(OSUtils::jsonString(member["id"],"0"));
@@ -566,7 +562,7 @@ unsigned int EmbeddedNetworkController::handleControlPlaneHttpGET(
 			std::set<std::string> networkIds;
 			{
 				Mutex::Lock _l(_db_m);
-				_db.filter("network/",120000,[&networkIds](const std::string &n,const json &obj) {
+				_db.filter("network/",[&networkIds](const std::string &n,const json &obj) {
 					if (n.length() == (16 + 8))
 						networkIds.insert(n.substr(8));
 					return true; // do not delete
@@ -641,7 +637,7 @@ unsigned int EmbeddedNetworkController::handleControlPlaneHttpPOST(
 					json member;
 					{
 						Mutex::Lock _l(_db_m);
-						member = _db.get("network",nwids,"member",Address(address).toString(),ZT_NETCONF_DB_CACHE_TTL);
+						member = _db.get("network",nwids,"member",Address(address).toString());
 					}
 					json origMember(member); // for detecting changes
 					_initMember(member);
@@ -667,7 +663,7 @@ unsigned int EmbeddedNetworkController::handleControlPlaneHttpPOST(
 								// Member is being de-authorized, so spray Revocation objects to all online members
 								if (!newAuth) {
 									_clearNetworkMemberInfoCache(nwid);
-									Revocation rev(_node->prng(),nwid,0,now,ZT_REVOCATION_FLAG_FAST_PROPAGATE,Address(address),Revocation::CREDENTIAL_TYPE_COM);
+									Revocation rev((uint32_t)_node->prng(),nwid,0,now,ZT_REVOCATION_FLAG_FAST_PROPAGATE,Address(address),Revocation::CREDENTIAL_TYPE_COM);
 									rev.sign(_signingId);
 									Mutex::Lock _l(_lastRequestTime_m);
 									for(std::map< std::pair<uint64_t,uint64_t>,uint64_t >::iterator i(_lastRequestTime.begin());i!=_lastRequestTime.end();++i) {
@@ -792,7 +788,7 @@ unsigned int EmbeddedNetworkController::handleControlPlaneHttpPOST(
 					test->timestamp = OSUtils::now();
 
 					if (_node) {
-						_node->circuitTestBegin(test,&(EmbeddedNetworkController::_circuitTestCallback));
+						_node->circuitTestBegin((void *)0,test,&(EmbeddedNetworkController::_circuitTestCallback));
 					} else {
 						_tests.pop_back();
 						return 500;
@@ -824,7 +820,7 @@ unsigned int EmbeddedNetworkController::handleControlPlaneHttpPOST(
 							uint64_t tryNwid = nwidPrefix | (nwidPostfix & 0xffffffULL);
 							if ((tryNwid & 0xffffffULL) == 0ULL) tryNwid |= 1ULL;
 							Utils::snprintf(nwids,sizeof(nwids),"%.16llx",(unsigned long long)tryNwid);
-							if (_db.get("network",nwids,ZT_NETCONF_DB_CACHE_TTL).size() <= 0) {
+							if (_db.get("network",nwids).size() <= 0) {
 								nwid = tryNwid;
 								break;
 							}
@@ -833,7 +829,7 @@ unsigned int EmbeddedNetworkController::handleControlPlaneHttpPOST(
 							return 503;
 					}
 
-					network = _db.get("network",nwids,ZT_NETCONF_DB_CACHE_TTL);
+					network = _db.get("network",nwids);
 				}
 				json origNetwork(network); // for detecting changes
 				_initNetwork(network);
@@ -1020,9 +1016,10 @@ unsigned int EmbeddedNetworkController::handleControlPlaneHttpPOST(
 									json ntag = json::object();
 									const uint64_t tagId = OSUtils::jsonInt(tag["id"],0ULL);
 									ntag["id"] = tagId;
-									if (tag.find("default") == tag.end())
-										ntag["default"] = json();
-									else ntag["default"] = OSUtils::jsonInt(tag["default"],0ULL);
+									json &dfl = tag["default"];
+									if (dfl.is_null())
+										ntag["default"] = dfl;
+									else ntag["default"] = OSUtils::jsonInt(dfl,0ULL);
 									ntags[tagId] = ntag;
 								}
 							}
@@ -1053,7 +1050,7 @@ unsigned int EmbeddedNetworkController::handleControlPlaneHttpPOST(
 					}
 
 					// Send an update to all members of the network
-					_db.filter((std::string("network/") + nwids + "/member/"),120000,[this,&now,&nwid](const std::string &n,const json &obj) {
+					_db.filter((std::string("network/") + nwids + "/member/"),[this,&now,&nwid](const std::string &n,const json &obj) {
 						_pushMemberUpdate(now,nwid,obj);
 						return true; // do not delete
 					});
@@ -1070,7 +1067,19 @@ unsigned int EmbeddedNetworkController::handleControlPlaneHttpPOST(
 
 		} // else 404
 
-	} // else 404
+	} else if (path[0] == "ping") {
+
+		json testRec;
+		const uint64_t now = OSUtils::now();
+		testRec["clock"] = now;
+		testRec["uptime"] = (now - _startTime);
+		testRec["content"] = b;
+		responseBody = OSUtils::jsonDump(testRec);
+		_db.writeRaw("pong",responseBody);
+		responseContentType = "application/json";
+		return 200;
+
+	}
 
 	return 404;
 }
@@ -1095,7 +1104,7 @@ unsigned int EmbeddedNetworkController::handleControlPlaneHttpDELETE(
 			json network;
 			{
 				Mutex::Lock _l(_db_m);
-				network = _db.get("network",nwids,ZT_NETCONF_DB_CACHE_TTL);
+				network = _db.get("network",nwids);
 			}
 			if (!network.size())
 				return 404;
@@ -1106,7 +1115,7 @@ unsigned int EmbeddedNetworkController::handleControlPlaneHttpDELETE(
 
 					Mutex::Lock _l(_db_m);
 
-					json member = _db.get("network",nwids,"member",Address(address).toString(),ZT_NETCONF_DB_CACHE_TTL);
+					json member = _db.get("network",nwids,"member",Address(address).toString());
 					_db.erase("network",nwids,"member",Address(address).toString());
 
 					if (!member.size())
@@ -1118,8 +1127,9 @@ unsigned int EmbeddedNetworkController::handleControlPlaneHttpDELETE(
 			} else {
 				Mutex::Lock _l(_db_m);
 
-				std::string pfx("network/"); pfx.append(nwids);
-				_db.filter(pfx,120000,[](const std::string &n,const json &obj) {
+				std::string pfx("network/");
+				pfx.append(nwids);
+				_db.filter(pfx,[](const std::string &n,const json &obj) {
 					return false; // delete
 				});
 
@@ -1246,8 +1256,8 @@ void EmbeddedNetworkController::_request(
 	json member;
 	{
 		Mutex::Lock _l(_db_m);
-		network = _db.get("network",nwids,ZT_NETCONF_DB_CACHE_TTL);
-		member = _db.get("network",nwids,"member",identity.address().toString(),ZT_NETCONF_DB_CACHE_TTL);
+		network = _db.get("network",nwids);
+		member = _db.get("network",nwids,"member",identity.address().toString());
 	}
 
 	if (!network.size()) {
@@ -1772,7 +1782,7 @@ void EmbeddedNetworkController::_getNetworkMemberInfo(uint64_t now,uint64_t nwid
 
 	{
 		Mutex::Lock _l(_db_m);
-		_db.filter(pfx,120000,[&nmi,&now](const std::string &n,const json &member) {
+		_db.filter(pfx,[&nmi,&now](const std::string &n,const json &member) {
 			try {
 				if (OSUtils::jsonBool(member["authorized"],false)) {
 					++nmi.authorizedMemberCount;
