@@ -207,14 +207,14 @@ bool CertificateOfMembership::sign(const Identity &with)
 	}
 }
 
-int CertificateOfMembership::verify(const RuntimeEnvironment *RR) const
+int CertificateOfMembership::verify(const RuntimeEnvironment *RR,void *tPtr) const
 {
 	if ((!_signedBy)||(_signedBy != Network::controllerFor(networkId()))||(_qualifierCount > ZT_NETWORK_COM_MAX_QUALIFIERS))
 		return -1;
 
-	const Identity id(RR->topology->getIdentity(_signedBy));
+	const Identity id(RR->topology->getIdentity(tPtr,_signedBy));
 	if (!id) {
-		RR->sw->requestWhois(_signedBy);
+		RR->sw->requestWhois(tPtr,_signedBy);
 		return 1;
 	}
 
