@@ -22,9 +22,9 @@
 
 namespace ZeroTier {
 
-bool Path::send(const RuntimeEnvironment *RR,const void *data,unsigned int len,uint64_t now)
+bool Path::send(const RuntimeEnvironment *RR,void *tPtr,const void *data,unsigned int len,uint64_t now)
 {
-	if (RR->node->putPacket(_localAddress,address(),data,len)) {
+	if (RR->node->putPacket(tPtr,_localAddress,address(),data,len)) {
 		_lastOut = now;
 		return true;
 	}
